/*
FILE:			clock.c
PROJECT:		Telemetry Buoy project
DESCRIPTION:	Functions for changing system or timer clock source and prescaling
DATE:			17.06.09
AUTHOR:			Stian O. Moen
*/

#include "CLOCK_functions.h"

//Init_clock will set the CPU to use the PLL as main oscillator, PLL is
//setup to provide 60Mhz from the external 12MHz crystal

void init_clock( void )
{
#ifdef EXTERNAL_CRYSTAL
	// Switch main clock to external oscillator 0 (crystal).
	pm_switch_to_osc0(&AVR32_PM, FOSC0, AVR32_PM_OSCCTRL0_STARTUP_4096_RCOSC);
#endif
#ifdef EXTERNAL_CLOCK
	//switch to external clock
	pm_enable_osc0_ext_clock(&AVR32_PM);
	pm_enable_clk0(&AVR32_PM, 0);
	pm_switch_to_clock(&AVR32_PM, AVR32_PM_MCSEL_OSC0);
#endif

	   volatile avr32_pm_t* pm = &AVR32_PM;

	   /* start PLL0 and switch main clock to PLL0 output */
	   local_start_pll0(pm);

}

/* Start PLL0, enable a generic clock with PLL0 output then switch main clock to PLL0 output.
   All calculations in this function suppose that the Osc0 frequency is 12MHz. */
void local_start_pll0(volatile avr32_pm_t* pm)
{
//  pm_switch_to_osc0(pm, FOSC0, OSC0_STARTUP);  // Switch main clock to Osc0.

  /* Setup PLL0 on Osc0, mul=9 ,no divisor, lockcount=16, ie. 12Mhzx10 = 120MHz output */
  /*void pm_pll_setup(volatile avr32_pm_t* pm,
                  unsigned int pll,
                  unsigned int mul,
                  unsigned int div,
                  unsigned int osc,
                  unsigned int lockcount) {
   */
  pm_pll_setup(pm,
               0,   // use PLL0
               9,   // MUL=7 in the formula
               1,   // DIV=1 in the formula
               0,   // Sel Osc0/PLL0 or Osc1/PLL1
               16); // lockcount in main clock for the PLL wait lock

  /*
   This function will set a PLL option.
   *pm Base address of the Power Manager (i.e. &AVR32_PM)
   pll PLL number 0
   pll_freq Set to 1 for VCO frequency range 80-180MHz, set to 0 for VCO frequency range 160-240Mhz.
   pll_div2 Divide the PLL output frequency by 2 (this settings does not change the FVCO value)
   pll_wbwdisable 1 Disable the Wide-Bandith Mode (Wide-Bandwith mode allow a faster startup time and out-of-lock time). 0 to enable the Wide-Bandith Mode.
  */
  /* PLL output VCO frequency is 120MHz. We divide it by 2 with the pll_div2=1. This enable to get later main clock to 60MHz */
  pm_pll_set_option(pm, 0, 1, 1, 0);

  /* Enable PLL0 */
  pm_pll_enable(pm,0);

  /* Wait for PLL0 locked */
  pm_wait_for_pll0_locked(pm) ;

  //setup PLL
  pm_gc_setup(pm,
              EXAMPLE_GCLK_ID,
              1,  // Use Osc (=0) or PLL (=1), here PLL
              0,  // Sel Osc0/PLL0 or Osc1/PLL1
              0,  // disable divisor
              0); // no divisor

  /* Enable Generic clock */
  pm_gc_enable(pm, EXAMPLE_GCLK_ID);

  pm_cksel(pm, 1, 0, 0, 0, 0, 0);

  // Set one wait-state (WS) for flash controller. 0 WS access is up to 30MHz for HSB/CPU clock.
  // As we want to have 60MHz on HSB/CPU clock, we need to set 1 WS on flash controller.
  flashc_set_wait_state(1);

  pm_switch_to_clock(pm, AVR32_PM_MCSEL_PLL0); /* Switch main clock to 60MHz */
}
