/*
 * RTC_functions.c
 *
 *  Created on: 02.jul.2009
 *      Author: Stian O. Moen
 */



#include "rtc.h"
#include "config.h"
#include "compiler.h"
#include "intc.h"
#include "RTC_functions.h"
#include "gpio.h"

//global variable containing the
time_variable_t time;


//RTC interrupt routine, updates a simplified real time clock.
//A clock with date and year can easily be implemented but is not
//needed for testing purpose.
__attribute__((__interrupt__))
void rtc_irq(void)
{
	//gpio_tgl_gpio_pin(AVR32_PIN_PA07);
  // Increment the seconds counter and handle other variables
  time.sec++;
  if (time.sec >= 60)
  {
	  time.sec = 0;
	  time.min++;
	  if (time.min >= 60)
	  {
		  time.min = 0;
		  time.hour++;
		  if (time.hour >= 24)
		  {
			  time.hour = 0;
			  time.day++;
		  }
	  }
  }

  // clear the interrupt flag
  rtc_clear_interrupt(&AVR32_RTC);
}

//the function will initialize the RTC module to use the external 32.000
//crystal as source, devide by 32 to count milliseconds and generate an
//interrupt every 1000ms to count seconds as well as setting this as the top
//for the timer.
void init_RTC(void)
{
	//initialize time to zero, this is only for testing. A real RTC should
	//always run and have functions to set the time. This can easily be
	//implemented when the user interface is defined.
	time.sec = 0;
	time.min = 0;
	time.hour = 0;
	time.day = 0;

	Disable_global_interrupt();

	//register RTC interrupt routine
	INTC_register_interrupt(&rtc_irq, AVR32_RTC_IRQ, AVR32_INTC_INT0);

	//initalize RTC to use 32kHz source
	rtc_init(&AVR32_RTC,/*0*/ RTC_OSC_32KHZ, RTC_PRESCALER);

	//set top to 1000 to generate interrupt each second.
	rtc_set_top_value(&AVR32_RTC, 1000);

	//Enable interrupt
	rtc_enable_interrupt(&AVR32_RTC);

	// Enable the RTC
	rtc_enable(&AVR32_RTC);

	Enable_global_interrupt();
}

//function returns time; intead of reading the global variable directly a
//this function assures that no writing is ongoing or will start during the
//read. This will prevent inconsistance between the time variables.
time_variable_t get_time(void)
{
	while(rtc_is_interrupt(&AVR32_RTC)); //wait until RTC interrupt is finished
	Disable_global_interrupt();
	time_variable_t temp = time;
	temp.millisec = (unsigned short)rtc_get_value(&AVR32_RTC);
	Enable_global_interrupt();
	return temp;
}

//calculates t1-t2. t1 must be larger than t2 to avoid negative time.
time_variable_t calculate_time_diff(time_variable_t t1, time_variable_t t2)
{
	time_variable_t temp;

	//calculate days
	temp.day = t1.day - t2.day;

	//calculate hours
	if(t1.hour - t2.hour >= 0)
	{
		temp.hour = t1.hour - t2.hour;
	}
	else
	{
		temp.hour = t1.hour + 24 - t2.hour;
		temp.day--;
	}

	//calculate minutes
	if(t1.min - t2.min >= 0)
	{
		temp.min = t1.min - t2.min;
	}
	else
	{
		temp.min = t1.min + 60 - t2.min;
		temp.hour--;
	}

	//calculate seconds
	if(t1.sec - t2.sec >= 0)
	{
		temp.sec = t1.sec - t2.sec;
	}
	else
	{
		temp.sec = t1.sec + 60 - t2.sec;
		temp.min--;
	}

	//calculate milliseconds
	if (t1.millisec - t2.millisec >= 0)
		temp.millisec = t1.millisec - t2.millisec;
	else
	{
		temp.millisec = t1.millisec + 1000 - t2.millisec;
		temp.sec--;
	}

	return temp;
}

