/*
 * SPI_functions.c
 *
 *  Created on: 30.jun.2009
 *      Author: Stian O. Moen
 */

#include "spi.h"
#include "compiler.h"
#include "config.h"
#include "SPI_functions.h"
#include "gpio.h"

volatile avr32_spi_t *spi = &AVR32_SPI;

unsigned char pot1_setting = 0;
unsigned char pot2_setting = 0;
unsigned char pot3_setting = 0;

//SPI options for setup with the dataflash
static const spi_options_t DATAFLASH_options =
{
	.reg = 3,
	.baudrate = 10000000,
	.bits = 8,
	.spck_delay = 0,
	.trans_delay = 0,
	.stay_act = 1,
	.spi_mode = 0,
	.modfdis = 1,
};

//SPI options for potentiometer updates.
static spi_options_t POTENTIOMETER_options =
{
	.reg = 0,
	.baudrate = 10000000,
	.bits = 8,
	.spck_delay = 5,
	.trans_delay = 5,
	.stay_act = FALSE,
	.spi_mode = 0,
	.modfdis = TRUE
};
//initialise to use potentiometer settings and set as master.
//The CS for the dataflash is set to high manually by the function
void init_SPI(void)
{
	static const gpio_map_t SPI_GPIO_MAP =
	  {
		{SPI_SCK_PIN, SPI_SCK_FUNCTION},
		{SPI_MISO_PIN, SPI_MISO_FUNCTION},
		{SPI_MOSI_PIN, SPI_MOSI_FUNCTION},
		{SPI_NPCS0_PIN, SPI_NPCS0_FUNCTION},
		{SPI_NPCS1_PIN, SPI_NPCS1_FUNCTION},
		{SPI_NPCS2_PIN, SPI_NPCS2_FUNCTION},
		{SPI_NPCS3_PIN, SPI_NPCS3_FUNCTION}
	  };

	  // Assign GPIO to SPI
	  gpio_enable_module(SPI_GPIO_MAP,
						 sizeof(SPI_GPIO_MAP) / sizeof(SPI_GPIO_MAP[0]));

	  //initialize as master
	  spi_initMaster(spi, &POTENTIOMETER_options);

	  // Set selection mode: variable_ps, pcs_decode, delay.
	  spi_selectionMode(spi, 0, 0, 0);

	  // Enable SPI.
	  spi_enable(spi);

	  //set manual chip select for dataflash to default high
	  gpio_set_gpio_pin(SPI_DATAFLASH_CS);

	  //initialize all potentiometers
	  short i;
	  for (i = 0; i < 4; i++)
	  {
		  POTENTIOMETER_options.reg = i;
		  spi_setupChipReg(spi, &POTENTIOMETER_options, F_CPU/2);
	  }

}

//Funtion sets the gain using the digital potentiometers
//This function will provide a maximum level throughout the
//cicuit by prioritizing the first gain stages.
void set_gain(U32 gain)
{
	unsigned short n1,n2,n3;    //potentiometer settings
	float g1, g2, g3; 			//gain of the induvidual stages

	//calculate gain for each gain stage
	if (gain <= GAIN1) //gain is only needed in stage 1
	{
		g1 = gain;
		g2 = 1;
		g3 = 1;
	}
	else if(gain <= GAIN1*GAIN2)
	{
		g1 = GAIN1;
		g2 = gain / GAIN1;
		g3 = 1;
	}
	else
	{
		g1 = GAIN1;
		g2 = GAIN2;
		g3 = (gain / GAIN1) / GAIN2;
	}

	//calculate potentiometer settings and ensure that overflow does not occur
	n1 = (g1 * POT_RES / GAIN1 + 0.5);
	if (n1 > 255) n1 = 255;
	n2 = (g2 * POT_RES / GAIN2 + 0.5);
	if (n2 > 255) n2 = 255;
	else if (n2 < 1) n2 = 1;
	n3 = (g3 * POT_RES / GAIN3 + 0.5);
	if (n3 > 255) n3 = 255;

	//set potentiometer gain only if different from last value (to reduce noise)
	if (n1 != pot1_setting)
		update_potentiometer(0, (char) n1);
	if (n2 != pot2_setting)
		update_potentiometer(1, (char) n2);
	if (n3 != pot3_setting)
		update_potentiometer(2, (char) n3);
}

//updates the given potentiometer with the raw input data
void update_potentiometer(unsigned char potentiometer, char data)
{
	spi_selectChip(spi, potentiometer);
	while(spi_writeEndCheck(spi) == 0);
	spi_write(spi, data);
	spi_unselectChip(spi, potentiometer);
}

//sets the threshold for the reception detector interrup with the raw input data
void set_threshold(unsigned char data)
{
	update_potentiometer(3, data);
}

//writes up-to one page to address zero. Different addresses can easily be implemented
//but is not needed for current measurement testing.
unsigned char store_data_in_dataflash(unsigned char *data, unsigned short size)
{
	//perform sanity check
	if(size > DATAFLASH_PAGE_SIZE)
		return 0;
	spi_setupChipReg(spi, &DATAFLASH_options, F_CPU/2);
	//The dataflash is not connected to a NPCS pin. Instead a gpio pin is set
	//manually to provide the chip select. To avoid that the SPI module
	//selects another device while programmeing the flash the NPCS3 pin is
	//rearranged to a gpio pin, thereby removing NPCS3 control for the SPI module
	//And the NPCS3 is selected when communicating with the serial flash.
	gpio_enable_gpio_pin(SPI_NPCS3_PIN);

	//set high to not select device before intentional select
	gpio_set_gpio_pin(SPI_NPCS3_PIN);
	spi_selectChip(spi, 3);
	//select dataflash
	gpio_clr_gpio_pin(SPI_DATAFLASH_CS);

	//check that SPI is not busy
	while(spi_writeEndCheck(spi) == 0);

	//write command "buffer1 write"
	spi_write(spi, CMD_WRITE_BUFFER_1);
	while(spi_writeEndCheck(spi) == 0);

	//write dont cares and buffer start index = 0
	spi_write(spi, 0);
	while(spi_writeEndCheck(spi) == 0);
	spi_write(spi, 0);
	while(spi_writeEndCheck(spi) == 0)
	;spi_write(spi, 0);
	while(spi_writeEndCheck(spi) == 0);

	//write data to buffer
	int i;
	for(i = 0; i < size; i++)
	{
		spi_write(spi, data[i]);
		while(spi_writeEndCheck(spi) == 0);
	}
	//signal end of command by setting CS high
	gpio_set_gpio_pin(SPI_DATAFLASH_CS);

	//send command "Buffer 1 to Main Memory Page Program with Built-in Erase"
	gpio_clr_gpio_pin(SPI_DATAFLASH_CS);
	spi_write(spi, CMD_BUFFER1_TO_MEM_WITH_ERASE);
	while(spi_writeEndCheck(spi) == 0);

	//write address bits and don't care. for testing the address is always 0
	spi_write(spi, 0);
	while(spi_writeEndCheck(spi) == 0);
	spi_write(spi, 0);
	while(spi_writeEndCheck(spi) == 0)
	;spi_write(spi, 0);
	while(spi_writeEndCheck(spi) == 0);

	//deselect serial flash
	gpio_set_gpio_pin(SPI_DATAFLASH_CS);

	//set options back to potentiometer settings
	POTENTIOMETER_options.reg = 3;
	spi_setupChipReg(spi, &POTENTIOMETER_options, F_CPU/2);

	//arrange NPCS3 to SPI module
	gpio_enable_module_pin(SPI_NPCS3_PIN, SPI_NPCS3_FUNCTION);

	return 1;
}

//reads up-to one page to address zero. Different addresses can easily be implemented
//but is not needed debug testing. This rutine is only used to test the correctness
//of the write routine.
unsigned char read_data_from_dataflash(unsigned char *data, unsigned short size)
{
	//perform sanity check
	if(size > DATAFLASH_PAGE_SIZE)
		return 0;
	spi_setupChipReg(spi, &DATAFLASH_options, F_CPU/2);
	//The dataflash is not connected to a NPCS pin. Instead a gpio pin is set
	//manually to provide the chip select. To avoid that the SPI module
	//selects another device while programmeing the flash the NPCS3 pin is
	//rearranged to a gpio pin, thereby removing NPCS3 control for the SPI module
	//And the NPCS3 is selected when communicating with the serial flash.
	gpio_enable_gpio_pin(SPI_NPCS3_PIN);

	//set high to not select device before intentional select
	gpio_set_gpio_pin(SPI_NPCS3_PIN);
	spi_selectChip(spi, 3);
	//select dataflash
	gpio_clr_gpio_pin(SPI_DATAFLASH_CS);

	//check that SPI is not busy
	while(spi_writeEndCheck(spi) == 0);

	//write command "Read page from main memory"
	spi_write(spi, CMD_MAIN_PAGE_READ);
	while(spi_writeEndCheck(spi) == 0);

	//write three address bytes and four dummy bytes according to datasheet
	unsigned short i;
	for(i=0; i < 7; i++)
	{
		spi_write(spi, 0);
		while(spi_writeEndCheck(spi) == 0);
	}

	//read data
	for(i = 0; i < size; i++)
	{
		//write dummy to receive data
		spi_write(spi,0);
		static unsigned short temp;
		spi_read(spi, &temp);
		data[i] = (unsigned char) temp;
	}


	//deselect serial flash
	gpio_set_gpio_pin(SPI_DATAFLASH_CS);

	//set options back to potentiometer settings
	POTENTIOMETER_options.reg = 3;
	spi_setupChipReg(spi, &POTENTIOMETER_options, F_CPU/2);

	//arrange NPCS3 to SPI module
	gpio_enable_module_pin(SPI_NPCS3_PIN, SPI_NPCS3_FUNCTION);
	return 1;
}
